use std::collections::HashSet;

use winit::event::{Event, ElementState, MouseButton};

use crate::canvas::canvas_frame::{Drawable};
use crate::drawables::rect::Rect;
use crate::drawables::sprite::Sprite;
use crate::util::vertex::VertexTypeContainer;
use crate::util::tr_event::{TrEvent, TrWindowEvent, TrUIEvent};

#[derive(Debug, Clone)]
pub struct Slider {
    handle: Rect,
    guide: Vec<Rect>,

    scaler: u32,
    position: (f32, f32),
    size: (f32, f32),
    value: u16,
}

impl Slider {
    pub fn new(size: (f32, f32), position: (f32, f32), value: u16) -> Slider {

        // render the guide first
        let red = (1.0, 0.0, 0.0, 0.0);
        let green = (0.0, 1.0, 0.0, 0.0);
        let blue = (0.0, 1.0, 1.0, 0.0);
        let rg = (1.0, 1.0, 0.0, 0.0);

        let left_guide_bar = Rect::new((position.0, position.1), (2.0, size.1), 1, red);
        let right_guide_bar = Rect::new((position.0 + size.0, position.1), (2.0, size.1), 1, blue);
        let line = Rect::new((position.0, position.1 - (size.1 / 2.0)), (size.0, 2.0), 1, green);

        let scale = value as f32 / u16::max_value() as f32;
        let handle = Rect::new((position.0 + (size.0 * scale), position.1), (15.0, size.1), 1, rg);

        Slider {
            handle: handle,
            guide: vec![left_guide_bar, right_guide_bar, line],
            scaler: 255,
            position,
            size,
            value,
        }
    }

    pub fn update(&mut self) {

        // render the guide first
        let red = (1.0, 0.0, 0.0, 0.0);
        let green = (0.0, 1.0, 0.0, 0.0);
        let blue = (0.0, 1.0, 1.0, 0.0);
        let rg = (1.0, 1.0, 0.0, 0.0);

        let left_guide_bar = Rect::new((self.position.0, self.position.1), (2.0, self.size.1), 1, red);
        let right_guide_bar = Rect::new((self.position.0 + self.size.0, self.position.1), (2.0, self.size.1), 1, blue);
        let line = Rect::new((self.position.0, self.position.1 - (self.size.1 / 2.0)), (self.size.0, 2.0), 1, green);

        let scale = self.value as f32 / u16::max_value() as f32;
        let handle = Rect::new((self.position.0 + (self.size.0 * scale), self.position.1), (15.0, self.size.1), 1, rg);

        self.handle = handle;
        self.guide = vec![left_guide_bar, right_guide_bar, line];
        self.scaler = 255;
    }
}

impl Drawable for Slider {
    fn render(&self,
              window_size: (u32, u32),
              position: (f32, f32),
              rotation: f32,
              size: (f32, f32),
              depth: f32,
    ) -> Vec<VertexTypeContainer> {
        let mut vertices = self.handle.render(
            window_size,
            position,
            rotation,
            size,
            depth,
        ).clone();

        vertices.extend_from_slice(
            self.guide.iter()
                .map(|x| x.render(window_size,
                                  position,
                                  rotation,
                                  size,
                                  depth,
                )).flatten()
                .collect::<Vec<VertexTypeContainer>>()
                .as_slice()
        );

        vertices.extend_from_slice(
            self.guide[0].render(window_size,
                                 position,
                                 rotation,
                                 size,
                                 depth,
            ).as_slice());
        vertices
    }

    // fn update<T>(&self, delta_time: f32) -> Vec<TrUIEvent<T>> {
    //     Vec::new()
    // }
    //
    // fn notify<Y, T>(&self, tr_event: Vec<TrEvent<Y>>, ui_events: Vec<TrUIEvent<T>>) -> Vec<TrUIEvent<T>> {
    //     match tr_event {
    //         TrEvent::WindowEvent { event: TrWindowEvent::MouseInput { device_id, state, button, modifiers }, .. } => {
    //
    //
    //             match button {
    //                 MouseButton::Left => {
    //                     if *state == ElementState::Pressed {
    //                         self.position.0 += 30.0;
    //                         self.value += 10;
    //                         self.update()
    //                     }
    //                 }
    //                 _ => {}
    //             }
    //         }
    //         _ => {}
    //     }
    //     Vec::new()
    // }
}
